unit CryptoKey;

interface

uses Windows, SysUtils, wcrypt2, CryptoWrapper, CryptoProvider, Classes;

type
  TCryptoKeyType = (cktSignature, cktExchange);
  TCryptoKey = class(TCryptoWrapper)
    fProvider: TCryptoProvider;
    fKeyType: TCryptoKeyType;
    fKey: HCRYPTKEY;
  private
    function GetPublicKey: string;
    procedure SetPublicKey(const Value: string);
    function GetKey(const aBlobType: DWord): string;
    procedure Crypt(const aEncrypt: Boolean; const aInStrm, aOutStrm: TStream); overload;
  public
    constructor Create(const aProvider: TCryptoProvider; const aKeyType: TCryptoKeyType);
    destructor Destroy; override;
    procedure GenerateNewKeys;
    procedure DestroyKeys;

    function GetKeyHandle: HCRYPTKEY;

    function Encrypt(const aPlain: string): string; overload;
    procedure Encrypt(const aPlain, aCipher: TStream); overload;
    function Decrypt(const aCipher: string): string; overload;
    procedure Decrypt(const aCipher, aPlain: TStream); overload;

    property PublicKey: string read GetPublicKey write SetPublicKey;
  end;


implementation

uses
  CryptoConsts, Wcrypt2Errors, EncdDecd;//, ZLib;

{ TCryptoUserKey }

constructor TCryptoKey.Create(const aProvider: TCryptoProvider; const aKeyType: TCryptoKeyType);
begin
  fProvider := aProvider;
  fKeyType := aKeyType;
  fKey := 0;
end;

destructor TCryptoKey.Destroy;
begin
  DestroyKeys;
  inherited;
end;

procedure TCryptoKey.DestroyKeys;
begin
  if (fKey > 0) then
  begin
    if (not CryptDestroyKey ( fKey )) then
      ECryptoException.RaiseLastCryptoError()
    else
      fKey := 0;
  end;
end;

function TCryptoKey.GetKeyHandle: HCRYPTKEY;
begin
  Result := fKey;
end;

procedure TCryptoKey.GenerateNewKeys;
begin
  DestroyKeys;
  if not CryptGenKey( fProvider.GetProviderHandle, C_CryptoKeyTypes[fKeyType], 0, @fKey ) then
    ECryptoException.RaiseLastCryptoError();
end;

function TCryptoKey.GetKey(const aBlobType: DWord): string;
var
  lKey: array[Byte] of Char;
  lKeyLen: Cardinal;
begin
  Result := '';
  if not CryptExportKey(fKey, 0, aBlobType, 0, @lKey[0], @lKeyLen) then
    ECryptoException.RaiseLastCryptoError;
  Result := EncodeBuffer(lKey, lKeyLen);
end;

function TCryptoKey.GetPublicKey: string;
begin
  Result := GetKey(PUBLICKEYBLOB);
end;

procedure TCryptoKey.SetPublicKey(const Value: string);
var
  lKey: array[Byte] of Char;
  lKeyLen: Cardinal;
begin
  lKeyLen := Decode2Buffer(value, lKey);
  if not CryptImportKey(fProvider.GetProviderHandle, @lKey[0], lKeyLen, 0, 0, @fKey) then
    ECryptoException.RaiseLastCryptoError;
end;

procedure TCryptoKey.Crypt(const aEncrypt: Boolean;
  const aInStrm, aOutStrm: TStream);
const
  C_CryptoBufferSize = 128;
var
  lBuffer: PByte;
  lLen: Cardinal;
  lEnd: Boolean;
begin
  Assert( Assigned( aInStrm ));
  Assert( Assigned( aOutStrm ));

  GetMem(lBuffer, C_CryptoBufferSize);
  try
    repeat
      lEnd := aInStrm.Position >= aInStrm.Size;

      if aEncrypt then
      begin
        // 64 bytes of Plaintext = 128 bytes of ciphertext 
        lLen := aInStrm.Read(lBuffer^, C_CryptoBufferSize div 2);
        if not CryptEncrypt(
          fKey,
          0,
          lEnd,
          0,
          lBuffer,
          @lLen,
          C_CryptoBufferSize) then
        ECryptoException.RaiseLastCryptoError
      end else begin
        lLen := aInStrm.Read(lBuffer^, C_CryptoBufferSize);
        if not CryptDecrypt(
          fKey,
          0,
          lEnd,
          0,
          lBuffer,
          @lLen) then
        ECryptoException.RaiseLastCryptoError
      end;

      aOutStrm.Write( lBuffer^, lLen );
    until lEnd;
  finally
    FreeMem(lBuffer, C_CryptoBufferSize);
  end;
end;

function TCryptoKey.Decrypt(const aCipher: string): string;
var
  lPlain, lCipher: TStringStream;
begin
  lCipher := TStringStream.Create(  DecodeString( aCipher ));
  try
    lPlain := TStringStream.Create( '' );
    try
      Crypt(False, lCipher, lPlain);
      Result := lPlain.DataString;
    finally
      lPlain.Free;
    end;
  finally
    lCipher.Free;
  end;
end;

procedure TCryptoKey.Decrypt(const aCipher, aPlain: TStream);
begin
  Crypt(True, aCipher, aPlain);
end;

procedure TCryptoKey.Encrypt(const aPlain, aCipher: TStream);
begin
  Crypt(True, aPlain, aCipher);
end;

function TCryptoKey.Encrypt(const aPlain: string): string;
var
  lPlain, lCipher: TStringStream;
begin
  lPlain := TStringStream.Create( aPlain );
  try
    lCipher := TStringStream.Create( '' );
    try
      Crypt(True, lPlain, lCipher);
      Result := EncodeString( lCipher.DataString );
    finally
      lCipher.Free;
    end;
  finally
    lPlain.Free;
  end;
end;

{
function TCryptoUserKey.LoadFromContext: boolean;
var
  lError: HRESULT;
begin
  Result := CryptGetUserKey( fProvider.GetProviderHandle, AT_SIGNATURE, @fSigKey );
  if not Result then
  begin
    lError := GetLastError;
    if lError <> NTE_NO_KEY then
      ECryptoException.RaiseCryptoError(lError);
  end;
end;
}

end.
